from dataclasses import field
from functools import partial
from pathlib import Path
from typing import Callable, List, Union
from typing_extensions import Self

from flax import struct
from jax import numpy as jnp
import librosa
import numpy as np
import soundfile

from .loudness import jit_integrated_loudness
from .resample import resample


@struct.dataclass
class SaliencyParams:
    """
    The parameters for saliency detection.

    Args:
        enabled (bool): Whether to enable saliency detection.
        num_tries (int): Maximum number of attempts to find a salient section of audio (default 8).
        loudness_cutoff (float): Minimum loudness cutoff in decibels for determining salient audio (default -40).
        search_function (Union[Callable, str]): The search function for determining the random offset. The default is
            ``SaliencyParams.search_uniform``. Another option is ``SaliencyParams.search_bias_early`` which gradually
            searches earlier in the file as more attempts are made.
    """

    enabled: bool = field(default=False)
    num_tries: int = 8
    loudness_cutoff: float = -40.0

    # Note: Although Union[Callable, str] would be a better type annotation, it doesn't work well with argbind
    search_function: str = "SaliencyParams.search_uniform"

    @staticmethod
    def search_uniform(
        rng: np.random.Generator,
        offset: float,
        duration: float,
        total_duration: float,
        attempt: int,
        max_attempts: int,
    ):
        lower_bound = max(0.0, offset)
        upper_bound = max(total_duration - duration, lower_bound)
        return rng.uniform(lower_bound, upper_bound)

    @staticmethod
    def search_bias_early(
        rng: np.random.Generator,
        offset: float,
        duration: float,
        total_duration: float,
        attempt: int,
        max_attempts: int,
    ):
        lower_bound = max(0.0, offset)
        upper_bound1 = max(total_duration - duration, lower_bound)
        # linearly interpolate the upper bound based on number of attempts so far
        alpha = attempt / (max_attempts - 1) if max_attempts > 1 else 0
        upper_bound2 = min(upper_bound1, lower_bound + duration)
        upper_bound = upper_bound1 * (1 - alpha) + upper_bound2 * alpha
        return rng.uniform(lower_bound, upper_bound)


_str_max_length = 256


@struct.dataclass
class AudioTree:
    """
    A `flax.struct.dataclass`_ for holding audio information including a waveform, sample rate, and metadata.

    The ``AudioTree`` class is inspired by Descript AudioTools's `AudioSignal`_.
        .. _AudioSignal: https://github.com/descriptinc/audiotools/blob/master/audiotools/core/audio_signal.py
        .. _flax.struct.dataclass: https://flax.readthedocs.io/en/latest/api_reference/flax.struct.html#flax.struct.dataclass

    Args:
        audio_data (jnp.ndarray): Audio waveform data in JAX numpy tensor shaped ``(Batch, Channels, Samples)``
        sample_rate (int): Sample rate of ``audio_data``, such as 44100 Hz.
        loudness (jnp.ndarray, optional): Loudness of the audio waveform in LUFs. You may not need to set this when initializing. Instead,
            use ``replace_loudness()`` to create a new AudioTree with ``loudness`` calculated.
        pitch (jnp.ndarray, optional): The MIDI pitch where 60 is middle C. The shape is ``(Batch,)``.
        velocity (jnp.ndarray, optional): The MIDI velocity between 0 and 127. The shape is ``(Batch,)``.
        duration (jnp.ndarray, optional): The duration of the audio waveform in seconds (like a note duration). The shape is ``(Batch,)``.
        codes (jnp.ndarray, optional): The neural audio codec tokens for the audio.
        latents (jnp.ndarray, optional): The latent representations of the audio.
        metadata (dict): Any extra metadata can be placed here.
    """

    audio_data: np.ndarray
    sample_rate: int = struct.field(pytree_node=False)
    loudness: np.ndarray = None
    pitch: np.ndarray = None
    velocity: np.ndarray = None
    duration: np.ndarray = None
    codes: np.ndarray = None
    latents: np.ndarray = None
    metadata: dict = struct.field(pytree_node=True, default_factory=dict)

    def replace_loudness(self) -> Self:
        """Replace ``loudness`` property with a JAX scalar."""
        loudness = jit_integrated_loudness(
            jnp.array(self.audio_data), self.sample_rate, zeros=512
        )
        return self.replace(loudness=loudness)

    @staticmethod
    def _encode_string(s: str) -> np.ndarray:
        # Convert string to list of ASCII values and pad with 0
        encoded = [ord(char) for char in s] + [0] * (_str_max_length - len(s))
        return np.array([encoded])  # [1, _str_max_length]

    @staticmethod
    def _decode_string(encoded_array: np.ndarray) -> str:
        # Convert list of integers to characters and join them into a string
        decoded = "".join([chr(val) for val in encoded_array if val != 0])
        return decoded

    @property
    def filepath(self) -> List[str]:
        """Return a list of filepaths assuming information exists in ``metadata['filepath']]``"""
        return [self._decode_string(data) for data in self.metadata["filepath"]]

    @classmethod
    def from_file(
        cls,
        audio_path: str,
        sample_rate: int = None,
        offset: float = 0.0,
        duration: float = None,
        mono: bool = False,
    ):
        """Create an AudioTree from an audio file path.

        Args:
            audio_path (str): Path to audio file.
            sample_rate (int, optional): Sample rate of audio data, such as 44100 Hz. If left as ``None``, the file's
                original sample rate will be used.
            offset (float, optional): Offset in seconds to audio data.
            duration (float, optional): Duration in seconds of audio data. The audio data will be trimmed or extended as
                necessary.
            mono (bool, optional): Whether to force the audio data to be single-channel.

        Returns:
            AudioTree: An instance of ``AudioTree``.
        """
        data, sr = librosa.load(
            audio_path, sr=sample_rate, offset=offset, duration=duration, mono=mono
        )
        assert sr == sample_rate
        if data.ndim == 1:
            data = data[None, None, :]  # Add batch and channel dimension
        elif data.ndim == 2:
            data = data[None, :, :]  # Add batch dimension

        if duration is not None and data.shape[-1] < round(duration * sample_rate):
            pad_right = round(duration * sample_rate) - data.shape[-1]
            data = np.pad(data, ((0, 0), (0, 0), (0, pad_right)))

        return cls(
            audio_data=data,
            sample_rate=sr,
            metadata={
                "filepath": cls._encode_string(audio_path),
                "offset": np.array([offset]),
                "duration": np.array([duration]),
            },
        )

    @classmethod
    def from_array(cls, audio_data: np.ndarray, sample_rate: int) -> Self:
        """Create an AudioTree from an audio array and a sample rate.

        Args:
            audio_data (np.ndarray): Audio data shaped ``(Samples)``, ``(Channels, Samples)``, or ``(Batch, Channels, Samples)``
            sample_rate (int): Sample rate of audio data, such as 44100 Hz.

        Returns:
            AudioTree: An instance of ``AudioTree``.
        """
        if audio_data.ndim == 1:
            audio_data = audio_data[None, None, :]  # Add batch and channel dimension
        elif audio_data.ndim == 2:
            audio_data = audio_data[None, :, :]  # Add batch dimension
        return cls(audio_data=audio_data, sample_rate=sample_rate)

    @classmethod
    def excerpt(
        cls,
        audio_path: str,
        rng: np.random.Generator,
        offset: float = 0.0,
        duration: float = None,
        search_function: Callable = None,
        **kwargs,
    ) -> Self:
        """Create an AudioTree from a random section of audio from a file path.

        Args:
            audio_path (str): Path to audio file.
            rng (np.random.Generator): Random number generator.
            offset (float, optional): Offset in seconds to audio data.
            duration (float, optional): Duration in seconds of audio data. The audio data will be trimmed or lengthened
                as necessary.
            search_function (Callable, optional): A function that determines the random offset.
            **kwargs: Keyword arguments passed to ``AudioTree.__init__``.

        Returns:
            AudioTree: An instance of ``AudioTree``.
        """
        assert duration is not None and duration > 0
        info = soundfile.info(audio_path)
        total_duration = info.duration  # seconds

        if search_function is None:
            search_function = partial(
                SaliencyParams.search_uniform, attempts=0, max_attempts=1
            )

        random_offset = search_function(rng, offset, duration, total_duration)

        audio_signal = cls.from_file(
            audio_path=audio_path, offset=random_offset, duration=duration, **kwargs
        )

        return audio_signal

    @classmethod
    def salient_excerpt(
        cls,
        audio_path: Union[str, Path],
        rng: np.random.Generator,
        saliency_params: SaliencyParams,
        **kwargs,
    ) -> Self:
        """Create an AudioTree from a salient section of audio from a file path.

        Args:
            audio_path (str): Path to audio file.
            rng (np.random.Generator): Random number generator such as ``np.random.default_rng(42)``.
            saliency_params (SaliencyParams): Saliency parameters to use to find a salient section.
            **kwargs: Keyword arguments passed to ``AudioTree.__init__``.

        Returns:
            AudioTree: An instance of ``AudioTree``.
        """
        assert (
            "offset" not in kwargs
        ), "``salient_excerpt`` cannot be used with kwarg ``offset``."
        assert (
            "duration" in kwargs
        ), "``salient_excerpt`` must be used with kwarg ``duration``."
        if (
            not saliency_params.enabled
            or saliency_params.loudness_cutoff is None
            or np.isnan(saliency_params.loudness_cutoff)
        ):
            excerpt = cls.excerpt(audio_path, rng=rng, **kwargs)
        else:
            loudness = -np.inf
            current_try = 0
            num_tries = saliency_params.num_tries
            if isinstance(saliency_params.search_function, str):
                _search_function = eval(saliency_params.search_function)
            else:
                _search_function = saliency_params.search_function
            while loudness <= saliency_params.loudness_cutoff:
                search_function = partial(
                    _search_function,
                    attempt=current_try,
                    max_attempts=num_tries,
                )
                new_excerpt = cls.excerpt(
                    audio_path, rng=rng, search_function=search_function, **kwargs
                ).replace_loudness()
                if current_try == 0 or new_excerpt.loudness > loudness:
                    excerpt = new_excerpt
                    loudness = new_excerpt.loudness
                current_try += 1
                if num_tries is not None and current_try >= num_tries:
                    break

        # todo: revisit whether casting to numpy here actually prevents any slowdown with grain.
        excerpt = excerpt.replace(
            audio_data=np.array(excerpt.audio_data), loudness=np.array(excerpt.loudness)
        )
        return excerpt

    def to_mono(self) -> Self:
        """Reduce the ``audio_data`` to mono.

        Returns:
            AudioTree: An instance of ``AudioTree``.
        """
        audio_data = self.audio_data.mean(axis=1, keepdims=True)
        return self.replace(audio_data=audio_data, loudness=None)

    def resample(
        self,
        sample_rate: int,
        zeros: int = 24,
        rolloff: float = 0.945,
        output_length: int = None,
        full: bool = False,
    ) -> Self:
        """
        Resample the AudioTree's ``audio_data`` to a new sample rate. The algorithm is a JAX port of ``ResampleFrac``
        from the PyTorch library `Julius`_.

        .. _Julius: https://github.com/adefossez/julius/blob/main/julius/resample.py

        Args:
            sample_rate (int): The new sample rate of audio data, such as 44100 Hz.
            zeros (int, optional): number of zero crossing to keep in the sinc filter.
            rolloff (float): use a lowpass filter that is ``rolloff * sample_rate / 2``,
                to ensure sufficient margin due to the imperfection of the FIR filter used.
                Lowering this value will reduce antialiasing, but will reduce some of the
                highest frequencies.
            output_length (None or int): This can be set to the desired output length (last dimension).
                Allowed values are between 0 and ``ceil(length * sample_rate / old_sr)``. When ``None`` (default) is
                specified, the floored output length will be used. In order to select the largest possible
                size, use the `full` argument.
            full (bool): return the longest possible output from the input. This can be useful
                if you chain resampling operations, and want to give the ``output_length`` only
                for the last one, while passing ``full=True`` to all the other ones.

        Returns:
            AudioTree: An instance of ``AudioTree``.
        """
        if sample_rate == self.sample_rate:
            return self
        audio_data = resample(
            self.audio_data,
            self.sample_rate,
            sample_rate,
            zeros=zeros,
            rolloff=rolloff,
            output_length=output_length,
            full=full,
        )
        return self.replace(
            audio_data=audio_data, sample_rate=sample_rate, loudness=None
        )



from .core import Identity
from .core import VolumeChange
from .core import VolumeNorm
from .core import RescaleAudio
from .core import InvertPhase
from .core import SwapStereo
from .core import CorruptPhase
from .core import ShiftPhase
from .core import Choose
from .core import NeuralAudioCodecEncodeTransform
from .core import NeuralLatentEncodeTransform
from .core import ReduceBatchTransform

__all__ = [
    "Identity",
    "VolumeChange",
    "VolumeNorm",
    "RescaleAudio",
    "InvertPhase",
    "SwapStereo",
    "CorruptPhase",
    "ShiftPhase",
    "Choose",
    "NeuralAudioCodecEncodeTransform",
    "NeuralLatentEncodeTransform",
    "ReduceBatchTransform",
]



from functools import partial
from typing import Any, Optional, Tuple, Union

import chex
import jax
from einops import rearrange
from jax import numpy as jnp
from jax import random
from jax.tree_util import DictKey

from audiotree import AudioTree

KeyLeafPairs = list[tuple[list[DictKey], Any]]


def stft(x: jnp.ndarray, frame_length=2048, hop_factor=0.25, window="hann"):

    batch_size, num_channels, audio_length = x.shape

    frame_step = int(frame_length * hop_factor)

    x = rearrange(x, "b c t -> (b c) t")

    # This probably uses less memory than the aux method, but it's definitely slower than aux.
    _, _, stft_data = jax.scipy.signal.stft(
        x,
        window=window,
        nperseg=frame_length,
        noverlap=(frame_length - frame_step),
    )
    stft_data = rearrange(stft_data, "(b c) nf nt -> b c nf nt", b=batch_size)

    return stft_data


def istft(
    stft_matrix: chex.Array,
    noverlap: int,
    window: Optional[Union[str, float, Tuple[str, float]]] = "hann",
    length: Optional[int] = None,
) -> chex.Array:
    _, reconstructed_signal = jax.scipy.signal.istft(
        stft_matrix,
        noverlap=noverlap,
        window=window,
    )

    # Trim or pad the output signal to the desired length
    if length is not None:
        if length > reconstructed_signal.shape[-1]:
            # Pad the signal if it is shorter than the desired length
            pad_width = length - reconstructed_signal.shape[-1]
            reconstructed_signal = jnp.pad(
                reconstructed_signal, ((0, 0), (0, 0), (0, pad_width)), mode="constant"
            )
        else:
            # Trim the signal if it is longer than the desired length
            reconstructed_signal = reconstructed_signal[..., :length]

    return reconstructed_signal


# def _where_with_p(rng, modified: jnp.ndarray, original: jnp.ndarray, p: float):
#     B = original.shape[0]
#
#     shape = (B,) + (1,) * (original.ndim - 1)
#     use_modified = random.uniform(rng, shape=shape, minval=0) < p
#
#     return jnp.where(use_modified, modified, original)


def _db2linear(decibels):
    return jnp.pow(10.0, decibels / 20.0)


def _volume_norm_transform(
    audio_tree: AudioTree, key: jax.Array, min_db: float, max_db: float
) -> AudioTree:

    audio_data = audio_tree.audio_data

    B = audio_data.shape[0]

    key, subkey = random.split(key)
    target_db = random.uniform(subkey, shape=(B,), minval=min_db, maxval=max_db)
    gain_db = target_db - audio_tree.loudness

    audio_data = audio_data * _db2linear(gain_db)[:, None, None]

    audio_tree = audio_tree.replace(audio_data=audio_data, loudness=target_db)

    return audio_tree


def _volume_change_transform(
    audio_tree: AudioTree, key: jax.Array, min_db: float, max_db: float
) -> (tuple)[AudioTree, jnp.ndarray]:

    audio_data = audio_tree.audio_data

    B = audio_data.shape[0]

    key, subkey = random.split(key)
    gain_db = random.uniform(subkey, shape=(B,), minval=min_db, maxval=max_db)

    audio_data = audio_data * _db2linear(gain_db)[:, None, None]

    audio_tree = audio_tree.replace(audio_data=audio_data)

    return audio_tree, gain_db


def _rescale_audio_transform(audio_tree: AudioTree) -> AudioTree:
    """Rescales audio to the range [-1, 1] only if the original audio exceeds those bounds. Useful if transforms have
    caused the audio to clip. It won't change the relative balance of multichannel audio.
    """
    audio_data = audio_tree.audio_data
    maxes = jnp.max(jnp.absolute(audio_data), axis=[-2, -1])
    maxes = jnp.expand_dims(maxes, [-2, -1])
    maxes = jnp.maximum(maxes, jnp.ones_like(maxes))
    audio_data = audio_data / maxes

    return audio_tree.replace(audio_data=audio_data)


def _invert_phase_audio_transform(audio_tree: AudioTree) -> AudioTree:
    audio_data = audio_tree.audio_data
    audio_data = -audio_data
    return audio_tree.replace(audio_data=audio_data)


def _swap_stereo_audio_transform(audio_tree: AudioTree) -> AudioTree:
    audio_data = audio_tree.audio_data
    audio_data = jnp.flip(audio_data, axis=1)
    return audio_tree.replace(audio_data=audio_data)


def _corrupt_phase(
    audio_tree: AudioTree,
    rng: jax.Array,
    amount: float,
    hop_factor: float = 0.5,
    frame_length: float = 2048,
    window: str = "hann",
) -> AudioTree:
    audio_data = audio_tree.audio_data
    B, C, length = audio_data.shape

    frame_step = int(frame_length * hop_factor)
    noverlap = frame_length - frame_step

    stft_fun = partial(
        stft, frame_length=frame_length, hop_factor=hop_factor, window=window
    )
    istft_fun = partial(istft, noverlap=noverlap, window=window, length=length)

    stft_data = stft_fun(audio_data)

    amt = random.uniform(
        rng, shape=stft_data.shape[:-1], minval=-jnp.pi * amount, maxval=jnp.pi * amount
    )

    stft_data = stft_data * jnp.expand_dims(jnp.exp(1j * amt), axis=-1)
    audio_data = istft_fun(stft_data)

    return audio_tree.replace(audio_data=audio_data)


def _shift_phase(
    audio_tree: AudioTree,
    key: jax.Array,
    amount: float,
    hop_factor: float = 0.5,
    frame_length: float = 2048,
    window: str = "hann",
) -> AudioTree:
    audio_data = audio_tree.audio_data
    B, C, length = audio_data.shape

    frame_step = int(frame_length * hop_factor)
    noverlap = frame_length - frame_step

    stft_fun = partial(
        stft, frame_length=frame_length, hop_factor=hop_factor, window=window
    )
    istft_fun = partial(istft, noverlap=noverlap, window=window, length=length)

    stft_data = stft_fun(audio_data)

    key, subkey = random.split(key)
    amt = random.uniform(
        subkey,
        shape=stft_data.shape[:-2],
        minval=-jnp.pi * amount,
        maxval=jnp.pi * amount,
    )

    stft_data = stft_data * jnp.expand_dims(jnp.exp(1j * amt), axis=(-2, -1))
    audio_data = istft_fun(stft_data)

    return audio_tree.replace(audio_data=audio_data)



from typing import Any, Callable, Dict, List, Union

from einops import rearrange
from grain import python as grain
import jax
from jax import numpy as jnp
import numpy as np

from audiotree import AudioTree
from audiotree.transforms.base import BaseRandomTransform, BaseMapTransform
from audiotree.transforms.helpers import (
    _volume_norm_transform,
    _volume_change_transform,
    _rescale_audio_transform,
    _invert_phase_audio_transform,
    _swap_stereo_audio_transform,
    _corrupt_phase,
    _shift_phase,
)


class Identity(BaseMapTransform):
    """
    A transform that returns each item without any modifications.

    .. code-block:: python

        @staticmethod
        def get_default_config() -> Dict[str, Any]:
            return {}
    """

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        return {}

    @staticmethod
    def _apply_transform(audio_tree: AudioTree) -> AudioTree:
        return audio_tree


class VolumeChange(BaseRandomTransform):
    """Change the volume by a uniformly randomly selected decibel value.

    .. code-block:: python

        @staticmethod
        def get_default_config() -> Dict[str, Any]:
            return {
                "min_db": 0,
                "max_db": 0,
            }
    """

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        return {
            "min_db": 0,
            "max_db": 0,
        }

    @staticmethod
    def _apply_transform(
        audio_tree: AudioTree, rng: jax.Array, min_db: float, max_db: float
    ) -> AudioTree:
        audio_tree, gain_db = _volume_change_transform(audio_tree, rng, min_db, max_db)
        if audio_tree.loudness is not None:
            audio_tree = audio_tree.replace(loudness=(audio_tree.loudness + gain_db))
        return audio_tree


class VolumeNorm(BaseRandomTransform):
    """Normalize the volume to a randomly selected loudness value specified in LUFS.

    .. code-block:: python

        @staticmethod
        def get_default_config() -> Dict[str, Any]:
            return {
                "min_db": 0,
                "max_db": 0,
            }
    """

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        return {
            "min_db": 0,
            "max_db": 0,
        }

    @staticmethod
    def _pre_transform(audio_tree: AudioTree) -> AudioTree:
        audio_tree = audio_tree.replace_loudness()
        return audio_tree

    @staticmethod
    def _apply_transform(
        audio_tree: AudioTree, rng: jax.Array, min_db: float, max_db: float
    ) -> AudioTree:
        return _volume_norm_transform(audio_tree, rng, min_db, max_db)


class RescaleAudio(BaseMapTransform):
    """
    Rescale the audio so that the largest absolute value is 1.0. If none of the values are outside the range
    ``[-1., 1.]``, then no transformation is applied.

    .. code-block:: python

        @staticmethod
        def get_default_config() -> Dict[str, Any]:
            return {}
    """

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        return {}

    @staticmethod
    def _apply_transform(audio_tree: AudioTree) -> AudioTree:
        return _rescale_audio_transform(audio_tree).replace(loudness=None)


class InvertPhase(BaseMapTransform):
    """
    Invert the phase of all channels of audio.

    .. code-block:: python

        @staticmethod
        def get_default_config() -> Dict[str, Any]:
            return {}
    """

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        return {}

    @staticmethod
    def _apply_transform(audio_tree: AudioTree) -> AudioTree:
        return _invert_phase_audio_transform(audio_tree)


class SwapStereo(BaseMapTransform):
    """Swap the channels of stereo audio.

    .. code-block:: python

        @staticmethod
        def get_default_config() -> Dict[str, Any]:
            return {}
    """

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        return {}

    @staticmethod
    def _apply_transform(audio_tree: AudioTree) -> AudioTree:
        return _swap_stereo_audio_transform(audio_tree)


class NeuralLatentEncodeTransform(BaseMapTransform):
    """Use a neural network to set the latents of the AudioTree.

    .. code-block:: python

        @staticmethod
        def get_default_config() -> Dict[str, Any]:
            return {}
    """

    def __init__(
        self,
        encoder_fn: Callable[[AudioTree], jnp.ndarray],
        config: Dict[str, Dict[str, Any]] = None,
        scope: Dict[str, Dict[str, Any]] = None,
        output_key: Union[str, Callable[[List[str]], str]] = None,
    ):
        """
        Initialize the base transform with a configuration, a flag for seed splitting, a probability, a scope, and an
        output key.

        Args:
            encoder_fn (Callable[[AudioTree], jnp.ndarray]): A function that takes an audio_tree and returns a latent
                sequence.
            config (Dict[str, Dict[str, Any]]): Configuration dictionary for the transform
            scope (Dict[str, Dict[str, Any]]): Dictionary indicating which modalities to apply the transform to
            output_key (Union[str, Callable[[List[str]], str]], optional): Key under which to store the transformed
                value. By default, the values will be transformed in-place.
        """
        self.encoder_fn = encoder_fn
        super().__init__(config, scope, output_key)

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        return {}

    def _apply_transform(self, audio_tree: AudioTree) -> AudioTree:
        if audio_tree.latents is None:
            latents = self.encoder_fn(audio_tree)
            audio_tree = audio_tree.replace(latents=latents)
        return audio_tree


class CorruptPhase(BaseRandomTransform):
    """
    Perform a phase corruption on the audio. The phase shift range is in the range ``[-pi * amount, pi * amount]``, and
    it's independently selected for each frequency in the STFT.

    .. code-block:: python

        @staticmethod
        def get_default_config() -> Dict[str, Any]:
            return {
                "amount": 1,
                "hop_factor": 0.5,
                "frame_length": 2048,
                "window": "hann",
            }
    """

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        return {
            "amount": 1,
            "hop_factor": 0.5,
            "frame_length": 2048,
            "window": "hann",
        }

    @staticmethod
    def _apply_transform(
        audio_tree: AudioTree,
        rng: jax.Array,
        amount: float,
        hop_factor: float,
        frame_length: int,
        window: str,
    ) -> AudioTree:
        return _corrupt_phase(audio_tree, rng, amount, hop_factor, frame_length, window)


class ShiftPhase(BaseRandomTransform):
    """
    Perform a phase shift on the audio. The phase shift range is in the range ``[-pi * amount, pi * amount]``.

    .. code-block:: python

        @staticmethod
        def get_default_config() -> Dict[str, Any]:
            return {
                "amount": 1,
                "hop_factor": 0.5,
                "frame_length": 2048,
                "window": "hann",
            }
    """

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        return {
            "amount": 1,
        }

    @staticmethod
    def _apply_transform(
        audio_tree: AudioTree, rng: jax.Array, amount: float
    ) -> AudioTree:
        return _shift_phase(audio_tree, rng, amount)


class Choose(grain.RandomMapTransform):
    """
    With probability ``prob``, choose ``c`` transform(s) among ``transforms`` with optional probability weights
    ``weights``.
    """

    def __init__(self, *transforms, c: int = 1, weights=None, prob: float = 1):

        if weights is not None:
            assert len(weights) == len(transforms)

        assert c <= len(transforms)

        self.c = c
        self.weights = weights
        assert 0 <= prob <= 1
        self.prob = prob

        self.transforms = transforms

    def random_map(self, element, rng: np.random.Generator):

        # Reference:
        # https://github.com/google/grain/blob/cbad82fddd4c5bd94b87d93d3f29849e8e59a501/grain/_src/python/data_loader.py#L481

        if rng.random() >= self.prob:
            return element

        transforms = rng.choice(
            self.transforms, size=(self.c,), replace=False, p=self.weights
        )

        for transform in transforms:

            if isinstance(transform, grain.MapTransform):
                element = transform.map(element)
            elif isinstance(transform, grain.RandomMapTransform):
                element = transform.random_map(element, rng)
            elif hasattr(transform, "np_random_map"):  # TfRandomMapTransform
                element = transform.np_random_map(element, rng)
            else:
                # If a `seed` is provided we treat the Callable as RandomMapTransform
                element = transform(element, rng)

        return element


class NeuralAudioCodecEncodeTransform(BaseMapTransform):
    """
    Use a neural audio codec such as `Descript Audio Codec (DAC) or EnCodec <https://github.com/DBraun/DAC-JAX>`_ to
    encode audio into tokens.

    .. code-block:: python

        @staticmethod
        def get_default_config() -> Dict[str, Any]:
            return {}
    """

    def __init__(
        self,
        encoder_fn: Callable[[AudioTree], jnp.ndarray],
        num_codebooks: int,
        config: Dict[str, Dict[str, Any]] = None,
        scope: Dict[str, Dict[str, Any]] = None,
        output_key: Union[str, Callable[[List[str]], str]] = None,
    ):
        """
        Initialize the base transform with a configuration, a flag for seed splitting, a probability, a scope, and an
        output key.

        Args:
            encoder_fn (Callable[[AudioTree], jnp.ndarray]): A function that takes audio shaped ``(B, C, T)`` and
                returns tokens shaped ``((B C), K, S)``, where ``T`` is length in samples, ``S`` is encoded sequence
                length, and ``K`` is number of codebooks.
            num_codebooks (int): The number of codebooks in the codec.
            config (Dict[str, Dict[str, Any]]): Configuration dictionary for the transform
            scope (Dict[str, Dict[str, Any]]): Dictionary indicating which modalities to apply the transform to
            output_key (Union[str, Callable[[List[str]], str]], optional): Key under which to store the transformed
                value. By default, the values will be transformed in-place.
        """
        self.encoder_fn = encoder_fn
        self.num_codebooks = num_codebooks
        super().__init__(config, scope, output_key)

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        return {}

    def _apply_transform(self, audio_tree: AudioTree) -> AudioTree:
        if audio_tree.codes is None:
            B, C, T = audio_tree.audio_data.shape
            codes = self.encoder_fn(audio_tree)
            codes = rearrange(
                codes, "(B C) K S -> B (K C) S", B=B, C=C, K=self.num_codebooks
            )
            audio_tree = audio_tree.replace(codes=codes)
        return audio_tree


class ReduceBatchTransform(grain.MapTransform):

    def __init__(self):
        pass

    def map(self, audio_tree: AudioTree) -> AudioTree:

        def f(leaf):
            if isinstance(leaf, (np.ndarray, jnp.ndarray)):
                if leaf.ndim > 1:
                    shape = leaf.shape
                    shape = (shape[0] * shape[1],) + shape[2:]
                    return leaf.reshape(shape)
            return leaf

        audio_tree = jax.tree.map(f, audio_tree)

        return audio_tree



# File under the MIT license, see https://github.com/adefossez/julius/LICENSE for details.
# Author: adefossez, 2020
"""
Differentiable, Pytorch based resampling.
Implementation of Julius O. Smith algorithm for resampling.
See https://ccrma.stanford.edu/~jos/resample/ for details.
This implementation is specially optimized for when new_sr / old_sr is a fraction
with a small numerator and denominator when removing the gcd (e.g. new_sr = 700, old_sr = 500).

Very similar to [bmcfee/resampy](https://github.com/bmcfee/resampy) except this implementation
is optimized for the case mentioned before, while resampy is slower but more general.

"""

import math
from typing import Optional

from jax import lax
from jax import numpy as jnp


def sinc(x):
    """Sinc function implemented in JAX."""
    return jnp.sinc(x / jnp.pi)


def resample(
    x: jnp.ndarray,
    old_sr: int,
    new_sr: int,
    zeros: int = 24,
    rolloff: float = 0.945,
    output_length: Optional[int] = None,
    full: bool = False,
) -> jnp.ndarray:
    """Resampling algorithm adapted from the pytorch library Julius:
    https://github.com/adefossez/julius/blob/main/julius/resample.py
    """

    if not isinstance(old_sr, int) or not isinstance(new_sr, int):
        raise ValueError("old_sr and new_sr should be integers")

    assert x.ndim == 3

    if new_sr == old_sr:
        return x

    batch_size, c, length = x.shape

    gcd = math.gcd(old_sr, new_sr)
    old_sr = old_sr // gcd
    new_sr = new_sr // gcd
    zeros = zeros
    rolloff = rolloff

    if old_sr == new_sr:
        return x

    sr = min(new_sr, old_sr) * rolloff
    _width = math.ceil(zeros * old_sr / sr)
    idx = jnp.arange(-_width, _width + old_sr)
    kernels = []
    for i in range(new_sr):
        t = (-i / new_sr + idx / old_sr) * sr
        t = jnp.clip(t, -zeros, zeros)
        t = t * jnp.pi
        window = jnp.cos(t / zeros / 2) ** 2
        kernel = sinc(t) * window
        kernel = kernel / kernel.sum()
        kernels.append(kernel)

    kernel = jnp.stack(kernels).reshape((new_sr, 1, -1))

    y = lax.conv_general_dilated(
        x,
        kernel,
        window_strides=(old_sr,),
        padding=((int(_width), int(_width + old_sr)),),
    )

    y = jnp.transpose(y, (0, 2, 1))
    y = jnp.reshape(y, x.shape[:-1] + (-1,))

    float_output_length = new_sr * length / old_sr
    max_output_length = jnp.ceil(float_output_length).astype(int)
    default_output_length = jnp.floor(float_output_length).astype(int)

    if output_length is None:
        applied_output_length = max_output_length if full else default_output_length
    elif output_length < 0 or output_length > max_output_length:
        raise ValueError(f"output_length must be between 0 and {max_output_length}")
    else:
        applied_output_length = output_length
        if full:
            raise ValueError("You cannot pass both full=True and output_length")

    return y[..., :applied_output_length]



from typing import Any, Callable, Dict, List, Union
import warnings

from grain.python import MapTransform, RandomMapTransform
import jax
from jax import random
from jax.tree import map_with_path
from jax.tree_util import DictKey
import numpy as np

from audiotree import AudioTree

KeyLeafPairs = list[tuple[list[DictKey], Any]]


def _get_config_val(
    config: KeyLeafPairs,
    lookup_path: list[DictKey],
    lookup_key: str,
    default: Any,
) -> Any:
    """
    Retrieve the configuration value for a given key and path.

    :param config: A list of key-leaf pairs from `tree_util.tree_flatten_with_path`.
    :param lookup_path: Path of the current element.
    :param lookup_key: Configuration key to look up
    :param default: Default value if key is not found
    :return: Configuration value
    """
    longest_len = 0
    matched_value = default
    for config_path, value in config:
        if config_path[-1].key == lookup_key:
            L = len(config_path)
            if config_path[:-1] == lookup_path[: L - 1] and L > longest_len:
                longest_len = L
                matched_value = value
    return matched_value


def _is_in_scope(
    scope: KeyLeafPairs,
    lookup_path: list[DictKey],
) -> bool:
    """
    Retrieve the configuration value for a given key and path.

    :param scope: A list of key-leaf pairs from `tree_util.tree_flatten_with_path`.
    :param lookup_path: Path of the current element.
    :return: Boolean indicating if the path is in scope
    """
    if not scope:
        return True
    matched_value = False
    for config_path, value in scope:
        L = len(config_path)
        if config_path[:-1] == lookup_path[: L - 1]:
            if not value:
                return False
            matched_value = True
    return matched_value


def merge_pytree(tree1, tree2):
    """Order matters!"""

    def is_leaf(leaf: dict):
        # todo: ask JAX experts about this
        if not isinstance(leaf, dict):
            return False
        values = list(leaf.values())
        while isinstance(values, list) and values:
            values = values[0]
        return not values or isinstance(values, AudioTree)

    def _combine(x, y):
        return {**x, **y}

    return jax.tree.map(_combine, tree1, tree2, is_leaf=is_leaf)


class BaseTransformMixIn:

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        """
        Get the default configuration for the transform.

        :return: Default configuration dictionary
        """
        raise NotImplementedError("Must be implemented in subclass")

    @staticmethod
    def _pre_transform(element):
        """
        Apply a transform that will occur regardless of ``prob``.
        """
        return element

    @staticmethod
    def _apply_transform(element, rng: jax.Array, **kwargs):
        """
        Apply the transformation to the given element.

        Args:
            element (Any): Element to be transformed.
            rng (jax.Array): jax.random.PRNGKey
            **kwargs: Additional type-annotated keyword arguments for the transformation.

        Returns:
            Any: The transformed element.
        """
        raise NotImplementedError("Must be implemented in subclass")

    def _post_process(self, old_tree, new_tree):
        if self.output_key is None:
            return new_tree

        assert isinstance(
            old_tree, dict
        ), "You specified `output_key`, but the transformed element is not a dict."

        def is_leaf(x):
            # todo: ask JAX experts about this
            if not isinstance(x, dict):
                return False
            values = list(x.values())
            while isinstance(values, list):
                values = values[0]
            return isinstance(values, AudioTree)  # or values is None

        # Use output_key to rename the nodes in the tree
        def rename_node(path: list[DictKey], leaf):
            full_path = [k.key for k in path]
            leaf = {
                self.output_key(full_path + [k]): v
                for k, v in leaf.items()
                if _is_in_scope(self.scope, tuple(path) + (DictKey(k),))
            }
            return leaf

        # Rename the deepest keys in the new tree using the `output_key` function.
        new_tree = map_with_path(rename_node, new_tree, is_leaf=is_leaf)

        # Merge the trees. Unfortunately, the order matters.
        new_tree = merge_pytree(new_tree, old_tree)
        return new_tree


class BaseRandomTransform(BaseTransformMixIn, RandomMapTransform):

    def __init__(
        self,
        config: Dict[str, Any] = None,
        split_seed: bool = True,
        prob: float = 1.0,
        scope: Dict[str, Any] = None,
        output_key: Union[str, Callable[[List[str]], str]] = None,
    ):
        """
        Initialize the base transform with a configuration, a flag for seed splitting, a probability, a scope, and an
        output key.

        Args:
            config (Dict[str, Any]): Configuration dictionary for the transform
            split_seed (bool, optional): Whether to split the seed for each leaf. Defaults to True.
            prob (float, optional): Probability of applying the transform. Defaults to 1.0.
            scope (Dict[str, Any], optional): Dictionary indicating which modalities to apply the transform to
            output_key (Union[str, Callable[[List[str]], str]], optional): Key under which to store the transformed
                value. By default, the values will be transformed in-place.
        """
        assert 0 <= prob <= 1
        self.default_config = self.get_default_config()
        self.config = jax.tree_util.tree_flatten_with_path(config or {})[0]
        self.split_seed = split_seed
        self.prob = prob
        self.scope = jax.tree_util.tree_flatten_with_path(scope or {})[0]
        if isinstance(output_key, str):
            # redefine it as a function
            self.output_key = lambda _: output_key
        else:
            self.output_key = output_key
        if output_key is not None and prob < 1.0:
            warnings.warn(
                "You have set a custom `output_key`, but `prob` is less than one. This may result in missing leaves."
            )

    def random_map(
        self, element: Any, rng: Union[np.random.Generator, jax.Array]
    ) -> Any:
        """
        Apply the random mapping to the given element using the provided seed.

        Args:
            element (Any): Input element to transform
            rng (jax.Array): jax.random.PRNGKey or numpy random generator
        Returns:
            Any: transformed element
        """

        if isinstance(rng, jax.Array):
            key = rng
        else:
            # todo: is there a better way to seed jax.random from this numpy random Generator?
            seed = rng.integers(2**63)
            key = random.PRNGKey(seed)

        def is_leaf(leaf):
            res = isinstance(leaf, AudioTree)  # todo: ask JAX experts about this
            return res

        def pre_transform_map_func(path: list[DictKey], leaf):
            if _is_in_scope(self.scope, path):
                transformed_leaf = self._pre_transform(leaf)
                return transformed_leaf
            return leaf

        def map_func(path: list[DictKey], leaf, rng: jax.Array, *config):
            if not is_leaf(leaf):
                return leaf
            if _is_in_scope(self.scope, path):
                transformed_leaf = self._apply_transform(leaf, rng, **config[0])
                return transformed_leaf
            elif self.output_key is not None:
                # The audiotree is not in scope, but output_key has been specified.
                # We drop the leaf and return None instead.
                # Then later when merging the new tree and the old tree, having None in the leaf helps us merge them
                # more easily.
                return None
            return leaf

        def map_use_default_config_val(path: list[DictKey], leaf):
            return {
                key: _get_config_val(self.config, path, key, default)
                for key, default in self.default_config.items()
            }

        element = map_with_path(pre_transform_map_func, element, is_leaf=is_leaf)

        treedef = jax.tree.flatten(element, is_leaf=is_leaf)[1]
        length = treedef.num_leaves
        subkeys = random.split(key, length) if self.split_seed else [key] * length
        subkeys = jax.tree.unflatten(treedef, subkeys)

        config = map_with_path(map_use_default_config_val, element, is_leaf=is_leaf)

        new_tree = map_with_path(map_func, element, subkeys, config, is_leaf=is_leaf)
        new_tree = self._post_process(element, new_tree)

        # Determine if we should apply the transform
        key, subkey = random.split(key)
        if self.prob == 1:
            return new_tree
        mask_new = random.uniform(subkey) < self.prob
        selected = jax.tree.map(
            (lambda x, y: jax.numpy.where(mask_new, x, y)), new_tree, element
        )
        return selected


class BaseMapTransform(BaseTransformMixIn, MapTransform):

    def __init__(
        self,
        config: Dict[str, Dict[str, Any]] = None,
        scope: Dict[str, Dict[str, Any]] = None,
        output_key: Union[str, Callable[[List[str]], str]] = None,
    ):
        """
        Initialize the base transform with a configuration, a flag for seed splitting, a probability, a scope, and an
        output key.

        Args:
            config (Dict[str, Dict[str, Any]]): Configuration dictionary for the transform
            scope (Dict[str, Dict[str, Any]]): Dictionary indicating which modalities to apply the transform to
            output_key (Union[str, Callable[[List[str]], str]], optional): Key under which to store the transformed
                value. By default, the values will be transformed in-place.
        """
        self.default_config = self.get_default_config()
        self.config = jax.tree_util.tree_flatten_with_path(config or {})[0]
        self.scope = jax.tree_util.tree_flatten_with_path(scope or {})[0]
        if isinstance(output_key, str):
            # redefine it as a function
            self.output_key = lambda _: output_key
        else:
            self.output_key = output_key

    def map(self, element: Any) -> Any:
        """
        Apply the random mapping to the given element.

        Args:
            element (Any): Input element to transform

        Returns:
            Any: transformed element
        """

        def is_leaf(leaf):
            res = isinstance(leaf, AudioTree)  # todo: ask JAX experts about this
            return res

        def pre_transform_map_func(path: list[DictKey], leaf):
            if not is_leaf(leaf):
                return leaf
            if _is_in_scope(self.scope, path):
                transformed_leaf = self._pre_transform(leaf)
                return transformed_leaf
            return leaf

        def map_func(path: list[DictKey], leaf, *config):
            if not is_leaf(leaf):
                return leaf
            if _is_in_scope(self.scope, path):
                transformed_leaf = self._apply_transform(leaf, **config[0])
                return transformed_leaf
            elif self.output_key is not None:
                # The audiotree is not in scope, but output_key has been specified.
                # We drop the leaf and return None instead.
                # Then later when merging the new tree and the old tree, having None in the leaf helps us merge them
                # more easily.
                return None
            return leaf

        def map_use_default_config_val(path: list[DictKey], leaf):
            return {
                key: _get_config_val(self.config, path, key, default)
                for key, default in self.default_config.items()
            }

        element = map_with_path(pre_transform_map_func, element, is_leaf=is_leaf)

        config = map_with_path(map_use_default_config_val, element, is_leaf=is_leaf)
        new_tree = map_with_path(map_func, element, config, is_leaf=is_leaf)
        return self._post_process(element, new_tree)



__version__ = "0.2.0"
__author__ = "David Braun"
# Effort-based versioning. Don't move the line above. It must be the first line due to `docs/source/conf.py`
from .core import AudioTree
from .core import SaliencyParams
from . import datasources
from . import transforms

__all__ = [
    "AudioTree",
    "SaliencyParams",
    "datasources",
    "transforms",
]



import glob
import math
import os
from random import Random
from typing import AnyStr, List, Mapping, SupportsIndex

from grain._src.python.dataset.transformations.mix import MixedIterDataset
from grain import python as grain
import numpy as np

from audiotree import AudioTree
from audiotree.core import SaliencyParams

_default_extensions = [".wav", ".flac"]


def _find_files_with_extensions(
    directory: str, extensions: List[str], max_depth=None, follow_symlinks=False
) -> list[AnyStr]:
    """
    Searches for files with specified extensions up to a maximum depth in the directory,
    without modifying dirs while iterating.

    Args:
        directory (str): The path to the directory to search.
        extensions (list): A list of file extensions to search for. Each extension should include a period.
        max_depth (int): The maximum depth to search for files.
        follow_symlinks (bool): Whether to follow symbolic links during the search.

    Returns:
        list (list[AnyStr]): A list of paths to files that match the extensions within the maximum depth.
    """
    matching_files = []
    extensions_set = {
        ext.lower() for ext in extensions
    }  # Normalize extensions to lowercase for matching

    # Expand environment variables and user home directory
    directory = os.path.expandvars(os.path.expanduser(directory))
    directory = os.path.abspath(directory)  # Ensure the directory path is absolute

    def recurse(current_dir, current_depth):
        if max_depth is not None and current_depth > max_depth:
            return
        with os.scandir(current_dir) as it:
            for entry in it:
                if entry.is_file(follow_symlinks=follow_symlinks) and any(
                    entry.name.lower().endswith(ext) for ext in extensions_set
                ):
                    matching_files.append(entry.path)
                elif entry.is_dir(follow_symlinks=follow_symlinks):
                    recurse(entry.path, current_depth + 1)

    recurse(directory, 0)
    return matching_files


class AudioDataSourceMixin:

    def load_audio(self, file_path, record_key: SupportsIndex) -> AudioTree:

        if self.saliency_params is not None and self.saliency_params.enabled:
            saliency_params = self.saliency_params
            return AudioTree.salient_excerpt(
                file_path,
                np.random.default_rng(int(record_key)),
                saliency_params=saliency_params,
                sample_rate=self.sample_rate,
                duration=self.duration,
                mono=self.mono,
            )

        return AudioTree.from_file(
            file_path,
            sample_rate=self.sample_rate,
            offset=0,
            duration=self.duration,
            mono=self.mono,
        )


class AudioDataSimpleSource(grain.RandomAccessDataSource, AudioDataSourceMixin):
    """A Data Source that aggregates all source files and weights them equally.

    Args:
        sources (Mapping[str, List[str]]): A dictionary mapping each source to a list of directories or glob
            expressions involving a file extension.
        num_records (int): The requested length of the data source.
        sample_rate (int): The requested sample rate of the audio.
        mono (bool): Whether to force the audio to be mono.
        duration (float): The requested duration of the audio.
        extensions (List[str]): A list of file extensions to search for. Each extension should include a period.
        saliency_params (SaliencyParams): Saliency parameters to use.
    """

    def __init__(
        self,
        sources: Mapping[str, List[str]],
        num_records: int = None,
        sample_rate: int = 44_100,
        mono: int = 1,
        duration: float = 1.0,
        extensions: List[str] = None,
        saliency_params: SaliencyParams = None,
    ):

        self.sample_rate = sample_rate
        self.mono = bool(mono)
        self.duration = duration
        if extensions is None:
            extensions = _default_extensions
        self.saliency_params = saliency_params

        filepaths = []
        for group_name, folders in sources.items():
            filepaths_in_group = []
            for _folder in folders:
                folder = os.path.expandvars(os.path.expanduser(_folder))
                if os.path.isdir(folder):
                    filepaths_in_group += _find_files_with_extensions(
                        folder, extensions=extensions
                    )
                else:
                    filepaths_in_group = list(glob.glob(folder, recursive=True))

            if filepaths_in_group:
                filepaths += filepaths_in_group
            else:
                raise RuntimeError(
                    f"Group '{group_name}' is empty. "
                    f"The number of specified folders in the group was {len(folders)}. "
                    f"The approved file extensions were {extensions}."
                )

        if num_records is not None:
            filepaths = filepaths[:num_records]

        self.filepaths = filepaths

        self._length = len(filepaths)
        assert self._length > 0

    def __len__(self) -> int:
        return self._length

    def __getitem__(self, record_key: SupportsIndex):
        file_path = self.filepaths[record_key]
        return self.load_audio(file_path, record_key)


class AudioDataBalancedSource(grain.RandomAccessDataSource, AudioDataSourceMixin):
    """A Data Source that equally weights multiple sources, where each source is a list of directories.

    Args:
        sources (Mapping[str, List[str]]): A dictionary mapping each source to a list of directories or glob
            expressions involving a file extension.
        num_records (int): The requested length of the data source.
        sample_rate (int): The requested sample rate of the audio.
        mono (bool): Whether to force the audio to be mono.
        duration (float): The requested duration of the audio.
        extensions (List[str]): A list of file extensions to search for. Each extension should include a period.
        saliency_params (SaliencyParams): Saliency parameters to use.
    """

    # todo: make this algorithm work if the user specifies weights for the groups.
    #  Right now the groups are balanced uniformly.
    #  Eventually the __init__ should just take a list of ``AudioDataSimpleSource`` and
    #  the corresponding weights.
    #  AudioDataBalancedDataset accomplishes this, but since it's an IterDataset it doesn't have all the features
    #  of a plain RandomAccessDataSource.

    def __init__(
        self,
        sources: Mapping[str, List[str]],
        num_records: int,
        sample_rate: int = 44_100,
        mono: int = 1,
        duration: float = 1.0,
        extensions: List[str] = None,
        saliency_params: SaliencyParams = None,
    ):

        self.sample_rate = sample_rate
        self.mono = bool(mono)
        self.duration = duration
        if extensions is None:
            extensions = _default_extensions
        self.saliency_params = saliency_params

        groups = []

        for group_name, folders in sources.items():
            filepaths = []
            for _folder in folders:
                folder = os.path.expandvars(os.path.expanduser(_folder))
                if os.path.isdir(os.path.expandvars(os.path.expanduser(folder))):
                    filepaths += _find_files_with_extensions(
                        folder, extensions=extensions
                    )
                else:
                    filepaths += list(glob.glob(folder))

            if filepaths:
                groups.append(filepaths)
            else:
                raise RuntimeError(
                    f"Group '{group_name}' is empty. "
                    f"The number of specified folders in the group was {len(folders)}. "
                    f"The approved file extensions were {extensions}."
                )

        self._num_groups = len(groups)
        self._length = num_records

        ideal_group_length = math.ceil(num_records / self._num_groups)
        seed = 0
        lengthened_groups = []
        for group in groups:
            num_loops = math.ceil(ideal_group_length / len(group))
            lengthened_group = []
            for _ in range(num_loops):
                copied = group.copy()
                Random(seed).shuffle(copied)
                seed += 1
                lengthened_group += copied
            lengthened_groups.append(lengthened_group)
        self._groups = lengthened_groups

        assert self._length > 0

    def __len__(self) -> int:
        return self._length

    def __getitem__(self, record_key: SupportsIndex):
        record_key = int(record_key)

        group_idx = record_key % self._num_groups
        idx = record_key // self._num_groups

        file_path = self._groups[group_idx][idx]

        return self.load_audio(file_path, record_key)


class AudioDataBalancedDataset(MixedIterDataset):
    """A Data Source that equally weights multiple sources, where each source is a list of directories.

    Args:
        sources (Mapping[str, List[str]]): A dictionary mapping each source to a list of directories or glob
            expressions involving a file extension.
        sample_rate (int): The requested sample rate of the audio.
        mono (bool): Whether to force the audio to be mono.
        duration (float): The requested duration of the audio.
        extensions (List[str]): A list of file extensions to search for. Each extension should include a period.
        saliency_params (SaliencyParams): Saliency parameters to use.
        weights (Mapping[str, float]): A dictionary mapping each source to its proportion in the dataset.
    """

    def __init__(
        self,
        sources: Mapping[str, List[str]],
        sample_rate: int = 44_100,
        mono: int = 1,
        duration: float = 1.0,
        extensions: List[str] = None,
        saliency_params: SaliencyParams = None,
        weights: Mapping[str, float] = None,
    ):
        self.sample_rate = sample_rate
        self.mono = bool(mono)
        self.duration = duration
        if extensions is None:
            extensions = _default_extensions
        self.saliency_params = saliency_params

        datasets = []

        seed = 0
        proportions = []
        for group_name, folders in sources.items():
            datasource = AudioDataSimpleSource(
                sources={group_name: folders},
                num_records=None,
                sample_rate=sample_rate,
                mono=mono,
                duration=duration,
                extensions=extensions,
                saliency_params=saliency_params,
            )
            dataset = (
                grain.MapDataset.source(datasource)
                .shuffle(seed=seed)
                .repeat()
                .to_iter_dataset()
            )
            seed += 1
            datasets.append(dataset)
            weight = 1.0
            if isinstance(weights, dict):
                weight = weights.get(group_name, 1.0)
            proportions.append(weight * 1000)

        super().__init__(datasets, proportions=proportions)



from .core import AudioDataSourceMixin
from .core import AudioDataBalancedSource
from .core import AudioDataBalancedDataset
from .core import AudioDataSimpleSource

__all__ = [
    "AudioDataSourceMixin",
    "AudioDataBalancedSource",
    "AudioDataBalancedDataset",
    "AudioDataSimpleSource",
]



import numpy as np
import jax.numpy as jnp

from audiotree.resample import resample


def _resample(y: np.ndarray, sr: int, new_sr: int, output_path: str):

    y = jnp.array(y)

    y = jnp.expand_dims(y, 0)
    y = y[:, :1, :]
    # print('y shape: ', y.shape)

    y = resample(y, old_sr=sr, new_sr=new_sr)
    # print('y shape: ', y.shape)
    y = y.squeeze(0).T
    y = np.array(y)

    # todo: use the torch version of julius and confirm the outputs match.
    # (DBraun did this manually once but didn't automate it.)
    # from scipy.io import wavfile
    # wavfile.write(output_path, new_sr, y)


# def test_resample_001(filepath='assets/60013__qubodup__whoosh.flac', new_sr=96_000):
#     import librosa
#     y, sr = librosa.load(filepath, sr=None, mono=False, duration=10)
#
#     _resample(y, sr, new_sr, "tmp_test_resample_001.wav")


def test_resample_002(new_sr=96_000):

    sr = 44100

    y = np.zeros((1, sr * 10))

    _resample(y, sr, new_sr, "tmp_test_resample_002.wav")



from functools import partial

from einops import rearrange
import jax
import jaxloudnorm as jln
from jax import numpy as jnp


@partial(jax.jit, static_argnames=("sample_rate", "zeros"))
def jit_integrated_loudness(data: jnp.ndarray, sample_rate: int, zeros: int):

    block_size = 0.4

    original_length = data.shape[-1]
    signal_duration = original_length / sample_rate

    if signal_duration < block_size:
        data = jnp.pad(
            data,
            pad_width=(
                (0, 0),
                (0, 0),
                (0, int(block_size * sample_rate) - original_length),
            ),
        )

    data = rearrange(data, "b c t -> b t c")
    meter = jln.Meter(sample_rate, block_size=block_size, use_fir=True, zeros=zeros)
    loudness = jax.vmap(meter.integrated_loudness)(data)

    loudness = jnp.where(
        jnp.isnan(loudness), jnp.full_like(loudness, -200), loudness
    )  # todo: -200 dB good default?

    return loudness
